(*<intr.pas>
 * Interrupt handling manager implementation for Turbo Pascal.
 * CopyLeft (c) since 1995 by PopolonY2k.
 *)

(**
  *
  * $Id: intr.pas 98 2015-08-21 01:28:40Z popolony2k $
  * $Author: popolony2k $
  * $Date: 2015-08-21 01:28:40 +0000 (Fri, 21 Aug 2015) $
  * $Revision: 98 $
  * $HeadURL: file:///svn/p/oldskooltech/code/msx/trunk/msxdos/pascal/intr.pas $
  *)

(*
 * This module depends on following include files (respect the order):
 * -
 *)


(**
  * Modules types and constants.
  *)
Type TInterruptAddress = Array[0..2] Of Byte;   { Interrupt address }



(**
  * Interrupt handler manager used internally to call the registered interrupt
  * user function.
  * WARNING: Do not call this function directly.
  *)
Procedure __Interrupt;
Begin
  InLine( $F3/          { DI          }
          $F5/          { PUSH AF     }
          $C5/          { PUSH BC     }
          $D5/          { PUSH DE     }
          $E5/          { PUSH HL     }
          $D9/          { EXX         }
          $F5/          { PUSH AF'    }
          $C5/          { PUSH BC'    }
          $D5/          { PUSH DE'    }
          $E5/          { PUSH HL'    }
          $DD/$E5/      { PUSH IX     }
          $FD/$E5/      { PUSH IY     }
          $21/$00/$00/  { LD HL, word }
          $E5/          { PUSH HL     }
          $CD/$00/$00/  { CALL nnnn   }
          $FD/$E1/      { POP IY      }
          $DD/$E1/      { POP IX      }
          $E1/          { POP HL      }
          $D1/          { POP DE      }
          $C1/          { POP BC      }
          $F1/          { POP AF      }
          $D9/          { EXX         }
          $E1/          { POP HL'     }
          $D1/          { POP DE'     }
          $C1/          { POP BC'     }
          $F1/          { POP AF'     }
          $C3/$00/$00   { JP nnnn     } );
End;

(**
  * Set a interrupt function to the interrupt vector.
  * @param nUserIntrFnAddr The user interrupt procedure address;
  * @param nUserParm The parameter passed by the user caller;
  * @param aOldIntr Reference to the address of the previous interrupt
  * procedure. The returned value will be useful to restore the old interrupt
  * function by calling the @see RestoreInterrupt function;
  *)
Procedure SetInterrupt( nUserIntrFnAddr,
                        nUserParm : Integer;
                        Var aOldIntr : TInterruptAddress );
Var
        nIntrFnAddr : Integer;
        aIM1Intr    : Array[0..2] Of Byte Absolute $0038;

Begin
  nIntrFnAddr := Addr( __Interrupt );
  Move( aIM1Intr, aOldIntr, SizeOf( aOldIntr ) );

  { Pass user parameter to the user interrupt handler }
  Mem[nIntrFnAddr+15] := Lo( nUserParm );
  Mem[nIntrFnAddr+16] := Hi( nUserParm );

  { Pass the user interrupt handler to the new interrupt handler manager }
  Mem[nIntrFnAddr+19] := Lo( nUserIntrFnAddr );
  Mem[nIntrFnAddr+20] := Hi( nUserIntrFnAddr );

  { Pass the old interrupt procedure to the new interrupt handler manager }
  Mem[nIntrFnAddr+34] := aIM1Intr[0];
  Mem[nIntrFnAddr+35] := aIM1Intr[1];
  Mem[nIntrFnAddr+36] := aIM1Intr[2];

  { Change the old interrupt handler by the new interrupt handler manager }
  InLine( $F3 );  { DI }
  aIM1Intr[1] := Lo( nIntrFnAddr );
  aIM1Intr[2] := Hi( nIntrFnAddr );
  InLine( $FB );  { EI }
End;

(**
  * Restore the interrupt address passed by parameter.
  * @param aIntr The interrupt address to be restored;
  *)
Procedure RestoreInterrupt( aIntr : TInterruptAddress );
Var
      aIM1Intr : Array[0..2] Of Byte Absolute $0038;

Begin
  InLine( $F3 );  { DI }
  Move( aIntr, aIM1Intr, SizeOf( aIntr ) );
  InLine( $FB );  { EI }
End;
