(**<popvgm.pas>
  * The Pop!Art VGM player engine for MSX computers.
  *
  * CopyLeft (c) since 1995 by PopolonY2k.
  *)

Program PopArt_VGM_Player;

(**
  *
  * $Id: popvgm.pas 98 2015-08-21 01:28:40Z popolony2k $
  * $Author: popolony2k $
  * $Date: 2015-08-21 01:28:40 +0000 (Fri, 21 Aug 2015) $
  * $Revision: 98 $
  * $HeadURL: file:///svn/p/oldskooltech/code/msx/trunk/msxdos/pascal/popvgm.pas $
  *)

(*
 * This module depends on folowing include files (respect the order):
 * - systypes.pas;
 * - system.pas;
 * - sleepb.pas;
 * - popmem.pas;
 * - types.pas;
 * - msxdos.pas;
 * - msxdos2.pas;
 * - msxbios.pas;
 * - dos2file.pas;
 * - sltsrch.pas;
 * - ay8910.pas;
 * - scc.pas;
 * - opl4.pas;
 * - y8950.pas;
 * - ym2413.pas;
 * - ym2151.pas;
 * - vgmtypes.pas;
 * - sndchips.pas;
 * - vgmfile.pas;
 * - vgmplay.pas;
 *)

{$v- c- u- a+ r- x+ }

{$i systypes.pas}
{$i system.pas}
{$i sleepb.pas}
{$i popmem.pas}
{$i types.pas}
{$i msxdos.pas}
{$i msxdos2.pas}
{$i msxbios.pas}
{$i dos2file.pas}
{$i sltsrch.pas}
{$i ay8910.pas}
{$i scc.pas}
{$i opl4.pas}
{$i y8950.pas}
{$i ym2413.pas}
{$i ym2151.pas}
{$i vgmtypes.pas}
{$i sndchips.pas}
{$i vgmfile.pas}
{$i vgmplay.pas}



(**
  * Engine's player constants.
  *)
Const
              ctEngineMinorVersion = 0;   { Engine's major version }
              ctEngineMajorVersion = 0;   { Engine's minor version }

(**
  * The command-line startup parameters.
  *)
Type TCmdLineParms = Record
  bFile                : Boolean;
  bHelp                : Boolean;
  bComputerFrequency   : Boolean;
  strFileName          : TFileName;
  computerFrequency    : THostInterruptTiming;
End;



(**
  * Parse the startup program parameters, returning the
  * result in @see TCmdLineParms structure;
  * @param parms The reference to parsed parameters command-line;
  *)
Procedure ParseCmdLineParms( Var parms : TCmdLineParms );
Var
      nCount,
      nParmCount   : Byte;
      strParm      : TFileName;

Begin
  nCount := 1;
  nParmCount := ParamCount;

  With parms Do
  Begin
    bComputerFrequency := False;
    bHelp := True;
    bFile := False;
    strFileName := '';
    computerFrequency := TimingUndefined;

    While( nCount <= nParmCount ) Do
    Begin
      strParm := ParamStr( nCount );

      { Process all possibles parameters values }
      If( bFile And ( strFileName = '' ) ) Then
      Begin
        strFileName := strParm;
        strParm := '';
      End
      Else
        If( bComputerFrequency And
            ( computerFrequency = TimingUndefined ) ) Then
        Begin
          If( strParm = '50' )  Then
            computerFrequency := Timing50Hz
          Else
            computerFrequency := Timing60Hz;

          strParm := '';
        End;

      If( ( Length( strParm ) >= 2 ) And ( strParm[1] = '-' ) ) Then
      Begin
        Case ( strParm[2] ) Of
          'h' : Begin                    { Help }
                  bHelp := True;
                  nCount := nParmCount;
                End;
          'f' : Begin                    { File name }
                  bFile := True;
                  bHelp := False;
                  strFileName := '';
                End;
          'c' : Begin                    { Computer frequency }
                  bComputerFrequency := True;
                End;
          Else Begin                     { Invalid parms }
                 bHelp  := True;
                 nCount := nParmCount;
               End;
        End;
      End;

      nCount := nCount + 1;
    End;
  End;
End;


(**
  * Show the engine's help.
  *)
Procedure ShowHelp;
Begin

  Write( 'Usage: ' );
  WriteLn( 'popvgm [-h][-c <50/60>] -f <file_name>' );
  WriteLn;
  WriteLn( '-h Show this help screen;' );
  WriteLn( '-c <50/60> Specify the computer''s frequency whose ' );
  WriteLn( '   the player will operates (50Hz or 60Hz) - Default 60Hz;' );
  WriteLn( '-f <file_name> The VGM file name (including complete path);' );
End;


(* Main block variables *)

Var
        vgmData     : TVGMData;
        chips       : TSoundChips;
        parms       : TCmdLineParms;

Begin
  WriteLn( 'Pop!Art vrs ', ctEngineMajorVersion, '.', ctEngineMinorVersion );
  WriteLn( 'CopyLeft (c) since 1995 by PopolonY2k.' );
  Write( 'Check newer versions of this software at ' );
  WriteLn( 'http://www.popolony2k.com.br' );
  WriteLn;

  ParseCmdLineParms( parms );

  If( Not parms.bHelp )  Then
  Begin
    InitChips( chips );

    chips.intrTiming := parms.computerFrequency;

    WriteLn( 'Opening VGM file [' + parms.strFileName + ']' );
    OpenVGM( parms.strFileName, vgmData {, chips} );

    (*
     * Check the opening status *)
    Case vgmData.status Of
      StateSuccessfullyLoaded :
        Begin
          If( parms.computerFrequency = Timing50Hz )  Then
            Write( '50Hz - ' )
          Else
            Write( '60Hz - ' );

          {WriteLn( 'Preprocessing' );
          ProcessVGM( vgmData, chips );}

          WriteLn( 'Press ESC key to exit.' );
          WriteLn( 'Playing' );
          PlayVGM( vgmData, chips );
        End;

      StateNotEnoughMemory :
        WriteLn( 'Not enough memory for this song' );

      StateInvalidFile,
      StateInvalidFileFormat,
      StateInvalidHeaderData :
        WriteLn( 'Invalid file' );
    End;

    CloseVGM( vgmData );
    ResetChips( chips );
  End
  Else
    ShowHelp;
End.
