/**
  * =========================================================================
  *  Copyright (C) since 1995 by Leidson CAF & PlanetaMessenger.org
  *
  *  This module is developed and maintained by PlanetaMessenger.org. 
  *  Specs, New and updated versions can be found in 
  *  http://www.planetamessenger.org
  *  If you want contact the Team please send a email to Project Manager
  *  Leidson Campos Alves Ferreira at leidson@planetamessenger.org
  *
  *  Copyright (C) since 1995 by PlanetaMessenger.org
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2 of the License, or
  *  (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  * ===========================================================================
  */

/**
  *
  * $Id: base.h 46 2011-03-08 13:23:21Z popolony2k $
  * $Author: popolony2k $
  * $Date: 2011-03-08 13:23:21 +0000 (Tue, 08 Mar 2011) $
  * $Revision: 46 $
  * $HeadURL: file:///svn/p/oldskooltech/code/msx/trunk/tools/cpp/sound/mml2msd/base.h $
  */

#ifndef __BASE_H__
#define __BASE_H__

#include <string>


/**
  * Channel definition for use in @see SetTempo().
  */
#define ALL_CHANNELS      -1


class IBase  {
  
  protected:

  /**
    * Must be overriden to implement a method to compile
    * needed information for IBase class based object.
    */  
  virtual bool Compile( void ) = 0;
  
  /**
    * Must be overriden to implement a method to link
    * needed information for IBase class based object.
    */  
  virtual bool Link( void ) = 0;

  public:
  
  /**
    * Virtual destructor.
    */
  virtual ~IBase( void )  {}
  
  /**
    * Must be overriden to implement a method to perform the open 
    * process to inherited class child;
    * The method must return:
    *  o false If the Open method has failed;
    *  o true If the Open method was succeded;
    */
  virtual bool Open( void ) = 0;

  /**
    * Must be overriden to implement a method to perform the close 
    * process to inherited class child;
    * The method must return:
    *  o false If the Close method has failed;
    *  o true If the Close method was succeded;
    */
  virtual bool Close( void ) = 0;

  /**
    * Must be overriden to implement a method to perform the Run 
    * process to inherited class child;
    * The method must return:
    *  o false If the Run method has failed;
    *  o true If the Run method was succeded;
    */
  virtual bool Run( void ) = 0;

  /**
    * Enable/Disable show information for verbose mode.
    * @param bEnableVerbose The new value for verbose information;
    */
  virtual void SetVerbose( bool bEnableVerbose ) = 0;

  /**
    * Return the verbose mode information.
    */
  virtual bool GetVerbose( void ) = 0;
  
  /**
    * Set the option for enable remarks on generated MSD MML
    * file.
    * @param bEnableRemarks The new settings for remarks;
    */
  virtual void SetRemarks( bool bEnableRemarks ) = 0;
  
  /**
    * Return if generated MSD MML file will be generated with
    * reamarks;
    */
  virtual bool GetRemarks( void ) = 0;

  /**
    * Set the flag to find music tempo based on input MML.
    * When this flag is set, the compiler will try to find 
    * the tempo from on source MML and apply it to all
    * other existing channels.
    * @param bFindTempo The new set to find tempo flag;
    */
  virtual void SetFindTempo( bool bFindTempo ) = 0;
  
  /**
    * Get the find tempo flag value assigned by the
    * @see SetFindTempo() method;
    */
  virtual bool GetFindTempo( void ) = 0;
  
  /**
    * Set the flag to remove all instruments os destination
    * MML.
    * @param bRemoveInstruments The new value to remove
    * instruments flag;
    */
  virtual void SetRemoveInstruments( bool bRemoveInstruments ) = 0;
  
  /**
    * Return the remove instruments flag assigned by
    * @see SetRemoveInstruments();
    */
  virtual bool GetRemoveInstruments( void ) = 0;

  /**
    * Add a custom MML command to a specified channel.
    * @param strCustomMML The custom MML to add to specified channel;
    * @param nChannel The Channel to add the custom MML;
    */
  virtual void AddCustomMML( std :: string strCustomMML, int nChannel = ALL_CHANNELS ) = 0;
  
  /**
    * Remove all custom MML commands added by
    * @see AddCustomMML().
    */
  virtual void RemovelAllCustomMML( void ) = 0;
};

#endif  // __BASE_H__
