/**
  * =========================================================================
  *  Copyright (C) since 1995 by Leidson CAF & PlanetaMessenger.org
  *
  *  This module is developed and maintained by PlanetaMessenger.org. 
  *  Specs, New and updated versions can be found in 
  *  http://www.planetamessenger.org
  *  If you want contact the Team please send a email to Project Manager
  *  Leidson Campos Alves Ferreira at leidson@planetamessenger.org
  *
  *  Copyright (C) since 1995 by PlanetaMessenger.org
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2 of the License, or
  *  (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  * ===========================================================================
  */

/**
  *
  * $Id: threeml.cpp 55 2011-03-26 20:07:52Z popolony2k $
  * $Author: popolony2k $
  * $Date: 2011-03-26 20:07:52 +0000 (Sat, 26 Mar 2011) $
  * $Revision: 55 $
  * $HeadURL: file:///svn/p/oldskooltech/code/msx/trunk/tools/cpp/sound/mml2msd/threeml.cpp $
  */

#include "threeml.h"
#include "threemltokens.h"
#include "msdtokens.h"
#include "stringutil.h"
#include <sstream>
#include <stdexcept>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>


/**
  * Internal module definitions.
  */
#define __MSX_MML_INC_TEMPO                  15
#define __MSD_CUSTOM_MML_VARIABLE            "C"
#define __MSD_CHANNEL_MML_VARIABLE           "CH"
#define __MSD_DEFAULT_TAG_FOR_EMPTY_CHANNEL  "@0"
#define __MSD_MAXIMUM_VOLUME_VALUE           "15"


// Private declarations
/**
  * Apply filter to Ampersand (&) character.
  * & after numbers is not possible;
  * & after dot (.) is not possible;
  * @param strToken The data to apply filter;
  */
void ThreeML :: AmpersandFilter( string &strToken )  {

  size_t    nCharPos = strToken.find( MSD_TOKEN_AMPERSAND );

  /*
   * Parse the value removing invalid & sequences.
   */
  while( nCharPos != string :: npos )  {
    if( nCharPos > 0 )  {
      char    nChar;

      try {
        nChar = strToken.at( nCharPos - 1 );
      } catch( out_of_range ex )  {
        nChar = MSD_TOKEN_SPACE;
      }

      /*
       * If preceding character is alpha or dot (.), remove
       * tie &
       */
      if( isalpha( nChar ) || ( nChar == MSD_TOKEN_DOT ) )
        strToken.replace( nCharPos, 1, 1, MSD_TOKEN_SPACE );
    }
 
    nCharPos = strToken.find( MSD_TOKEN_AMPERSAND, ++nCharPos );
  }
}

/**
  * Combine two quarter notes together.
  * @param strToken The MML data to filter;
  * Thanks to The Ultimate PPMCK MML reference.
  * http://woolyss.com/chipmusic/chipmusic-mml/ppmck_guide.php
  * and to SiOPMML reference.
  * http://mmltalks.appspot.com/document/siopm_mml_ref_05_e.html
  */
void ThreeML :: CombineQuarterNotesFilter( string &strToken )  {

  size_t    nCharPos = strToken.find( MSD_TOKEN_AMPERSAND );

  /*
   * Parse the value returning half of one note.
   * According Ultimate PPMCK MML reference.
   */
  while( nCharPos != string :: npos )  {
    if( nCharPos > 1 )  {
      char    nFirstDigit;
      char    nFirstNote;
      char    nSecondDigit;
      char    nSecondNote;

      try  {
        nFirstDigit = strToken.at( nCharPos - 1 );
      } catch( out_of_range ex )  {
        nFirstDigit = MSD_TOKEN_SPACE;
      }

      try  {
        nFirstNote = strToken.at( nCharPos - 2 );
      } catch( out_of_range ex )  {
        nFirstNote = MSD_TOKEN_1;
      }

      try  {
        nSecondDigit = strToken.at( nCharPos + 2 );
      } catch( out_of_range ex )  {
        nSecondNote = MSD_TOKEN_SPACE;
      }

      try  {
        nSecondNote = strToken.at( nCharPos + 1 );
      } catch( out_of_range ex )  {
        nSecondNote = MSD_TOKEN_1;
      }

      /*
       * Check notes and make half note or just clean it.
       */
      if( isdigit( nFirstDigit )        &&
          isalpha( nFirstNote )         && 
          isdigit( nSecondDigit )       &&
          isalpha( nSecondNote )        &&
          ( nFirstNote == nSecondNote ) && 
          ( nFirstDigit == nSecondDigit ) )  {

        char    szNoteDigit[6] = { 0, 0, 0, 0, 0, 0 };
        int     nDigit;

        szNoteDigit[0] = nFirstDigit;
        nDigit = ( nFirstDigit == MSD_TOKEN_1 ? atoi( szNoteDigit ) : ( atoi( szNoteDigit ) / 2 ) );
        sprintf( szNoteDigit, "%c%d%c%c%c", nFirstNote, nDigit, MSD_TOKEN_SPACE, MSD_TOKEN_SPACE, MSD_TOKEN_SPACE );
        strToken.replace( nCharPos - 2, strlen( szNoteDigit ), szNoteDigit );
      }
      else
        strToken.replace( nCharPos, 1, 1, MSD_TOKEN_SPACE );
    }
    else
      strToken.replace( nCharPos, 1, 1, MSD_TOKEN_SPACE );

    nCharPos = strToken.find( MSD_TOKEN_AMPERSAND, ++nCharPos );
  }
}

/**
  * Apply filter to Less-than (<) character.
  * Two << cannot be in sequence;
  * @param strToken The data to apply filter;
  */
void ThreeML :: OctaveLowerFilter( string &strToken )  {

  size_t    nCharPos = strToken.find( MSD_TOKEN_LESS_THAN );

  /*
   * Parse the value removing invalid & sequences.
   */
  while( nCharPos != string :: npos )  {
    char    nLessThan;


    try  {
      nLessThan = strToken.at( nCharPos + 1 );
    } catch( out_of_range ex )  {
      nLessThan = MSD_TOKEN_SPACE;
    }

    /*
     * If next character is a Less-Than '<' , remove
     * it.
     */
    if( nLessThan == MSD_TOKEN_LESS_THAN )
      strToken.replace( nCharPos, 1, 1, MSD_TOKEN_SPACE );

    nCharPos = strToken.find( MSD_TOKEN_LESS_THAN, ++nCharPos );
  }
}

/**
  * Apply filter to Dot (.) character.
  * . after L(n) is not possible;
  * @param strToken The data to apply filter;
  */
void ThreeML :: DotFilter( string &strToken )  {

  size_t    nCharPos = strToken.find( MSD_TOKEN_DOT );

  /*
   * Parse the value removing invalid & sequences.
   */
  while( nCharPos != string :: npos )  {
    if( nCharPos > 1 )  {
      char    nNote;
      char    nNoteDigit;
      
      
      try  {
        nNote = strToken.at( nCharPos - 2 );
      } catch( out_of_range ex )  {
        nNote = MSD_TOKEN_1;
      }

      try  {
        nNoteDigit = strToken.at( nCharPos - 1 );
      } catch( out_of_range ex )  {
        nNoteDigit = MSD_TOKEN_SPACE;
      }

      /*
       * Check if L MML command is followed by 2 digits.
       */
      if( isdigit( nNote ) )  {
        try  {
          nNote = strToken.at( nCharPos - 3 );
        } catch( out_of_range ex )  {
          nNote = MSD_TOKEN_1;
        }
      }

      /*
       * If preceding character is alpha or dot (.), remove
       * tie &
       */
      if( isalpha( nNote )      && 
          isdigit( nNoteDigit ) &&
          ( toupper( nNote ) == MSD_TOKEN_L ) )
        strToken.replace( nCharPos, 1, 1, MSD_TOKEN_SPACE );
    }
 
    nCharPos = strToken.find( MSD_TOKEN_DOT, ++nCharPos );
  }
}

/**
  * Apply filter to Tilde (~) character, removing it;
  * @param strToken The data to apply filter;
  */
void ThreeML :: TildeFilter( string &strToken )  {

  size_t    nCharPos = strToken.find( MSD_TOKEN_TILDE );

  /*
   * Remove the character ~(n) command.
   */
  while( nCharPos != string :: npos )  {
    if( nCharPos > 1 )  {
      char    nSecondDigit;

      try  {
        nSecondDigit = strToken.at( nCharPos + 2 );
      } catch( out_of_range ex )  {
        nSecondDigit = MSD_TOKEN_SPACE;
      }

      if( isdigit( nSecondDigit ) )
        strToken.replace( nCharPos, 1, 3, MSD_TOKEN_SPACE );
      else
        strToken.replace( nCharPos, 1, 2, MSD_TOKEN_SPACE );
    }

    nCharPos = strToken.find( MSD_TOKEN_TILDE );
  }
}

/**
  * Filter for invalid volume MML tag. Some MML sequences
  * is higher than MSD (MusicA) supports, then this will
  * be updated for maximum value accepted by MSD format (15).
  * @param strToken The data to apply filter;
  */
void ThreeML :: InvalidVolumeFilter( string &strToken )  {

  size_t    nCharPos = strToken.find( MSD_TOKEN_V );

  /*
   * Parse the value removing invalid & sequences.
   */
  while( nCharPos != string :: npos )  {
    char    nFirstDigit;
    char    nSecondDigit;
    char    nThirdDigit;
    char    szDigit[3] = { '0', '0', 0 };

    try  {
      nFirstDigit  = strToken.at( nCharPos + 1 );
    } catch( out_of_range ex )  {
      nFirstDigit = MSD_TOKEN_SPACE;
    }

    try  {
      nSecondDigit = strToken.at( nCharPos + 2 );
    } catch( out_of_range ex )  {
      nSecondDigit = MSD_TOKEN_SPACE;
    }
    
    try  {
      nThirdDigit = strToken.at( nCharPos + 3 );
    } catch( out_of_range ex )  {
      nThirdDigit = MSD_TOKEN_SPACE;
    }

    /*
     * Check if V MML command is followed by 2 digits.
     */
    if( isdigit( nSecondDigit ) && isdigit( nThirdDigit ) )  {
      sprintf( szDigit, "%c%c%c", nFirstDigit, nSecondDigit, nThirdDigit );
      strToken.replace( nCharPos + 3, 1, 1, MSD_TOKEN_SPACE );
    }
    else
      if( isdigit( nSecondDigit ) )
        sprintf( szDigit, "%c%c", nFirstDigit, nSecondDigit );

    if( atoi( szDigit ) > MSD_MML_MAX_VOLUME_VALUE )
      strToken.replace( nCharPos + 1, 2, __MSD_MAXIMUM_VOLUME_VALUE );

    nCharPos = strToken.find( MSD_TOKEN_V, ++nCharPos );
  }
}

/**
  * Filter all instruments of token, removing it from
  * sequence.
  * @param strToken The data to apply filter;
  */
void ThreeML :: InstrumentsFilter( string &strToken )  {

  size_t    nCharPos = strToken.find( MSD_TOKEN_AT );

  /*
   * Parse the value removing instruments.
   */
  while( nCharPos != string :: npos )  {
    char    nFirstDigit;
    char    nSecondDigit;
    char    nThirdDigit;

    try  {
      nFirstDigit = strToken.at( nCharPos + 1 );
    } catch( out_of_range ex )  {
      nFirstDigit = MSD_TOKEN_SPACE;
    }

    try  {
      nSecondDigit = strToken.at( nCharPos + 2 );
    } catch( out_of_range ex )  {
      nSecondDigit = MSD_TOKEN_SPACE;
    }
    
    try  {
      nThirdDigit = strToken.at( nCharPos + 3 );
    } catch( out_of_range ex )  {
      nThirdDigit = MSD_TOKEN_SPACE;
    }

    if( isdigit( nFirstDigit ) && isdigit( nSecondDigit ) && isdigit( nThirdDigit ) )
      strToken.replace( nCharPos, 1, 4, MSD_TOKEN_SPACE );
    else  {
      if( isdigit( nFirstDigit ) && isdigit( nSecondDigit ) )
        strToken.replace( nCharPos, 1, 3, MSD_TOKEN_SPACE );
      else  {
        if( isdigit( nFirstDigit ) )
          strToken.replace( nCharPos, 1, 2, MSD_TOKEN_SPACE );
        else
          strToken.replace( nCharPos, 1, 2, MSD_TOKEN_SPACE );
      }
    }

    nCharPos = strToken.find( MSD_TOKEN_AT, ++nCharPos );
  }
}

/**
  * Apply all pre filters to compiled MML.
  */
void ThreeML :: PreFilters( void )  {
  
  for( TokenVectorMap :: iterator itItems = compiledMML.begin(); itItems != compiledMML.end(); itItems++ )  {
    TokenVector    *pTokenVector = itItems -> second;
    
    if( pTokenVector )  {

      string      strValue( "" );

      /*
       * Must ignore the settings and 3ML Extension keys.
       */
      if( ( itItems -> first -> strKey.compare( THREEML_TOKEN_SETTINGS_SECTION_KEY ) == 0 ) ||
          ( itItems -> first -> strKey.compare( THREEML_TOKEN_3MLE_EXTENSION_KEY ) == 0 ) )
        continue;
 
      for( TokenVector :: iterator itTokenItems = pTokenVector -> begin(); itTokenItems != pTokenVector -> end(); itTokenItems++ )  {
        struct   stToken    *pToken = *itTokenItems;

        strValue += pToken -> strValue;
      }

      ::remove( strValue, MSD_TOKEN_SPACE );

      if( GetFindTempo() )
        FindTempo( strValue );
    }
  }
}

/**
  * Find the tempo on MML and add to all channels as custom
  * MML.
  * This feature is added according value passed by SetFindTempo().
  * @param strToken The token to check;
  */
void ThreeML :: FindTempo( string strToken )  {

  if( !bTempoFound )  {
    size_t    nCharPos = strToken.find( MSD_TOKEN_T );

    if( nCharPos != string :: npos )  {
      char    nFirstDigit;
      char    nSecondDigit;
      char    nThirdDigit;
      size_t  nLen = 2;

      try  {
        nFirstDigit = strToken.at( nCharPos + 1 );
      } catch( out_of_range ex )  {
        nFirstDigit = MSD_TOKEN_SPACE;
      }

      try  {
        nSecondDigit = strToken.at( nCharPos + 2 );
      } catch( out_of_range ex )  {
        nSecondDigit = MSD_TOKEN_SPACE;
      }

      try  {
        nThirdDigit = strToken.at( nCharPos + 3 );
      } catch( out_of_range ex )  {
        nThirdDigit = MSD_TOKEN_SPACE;
      }

      if( isdigit( nThirdDigit ) )
        nLen+=2;
      else
        if( isdigit( nSecondDigit ) )
          nLen+=1;

      AddCustomMML( strToken.substr( nCharPos, nLen ), ALL_CHANNELS );
      bTempoFound = true;
    }
  }
}

// Protected declarations
/**
  * Check input data types.
  * @param strToken Data to be checked;
  */
TokenType ThreeML :: CheckToken( string strToken )  {
  
  string    strTmp( strToken );
  
  // Check Token
  if( ( strTmp.find( THREEML_TOKEN_OPEN_BRACKET ) != string :: npos ) && ( strTmp.find( THREEML_TOKEN_CLOSE_BRACKET ) != string :: npos ) )
    return TOKEN_TYPE_SECTION;

  /*
   * Check for remarks. Removes all comments.
   */
  // One line remarks
  if( strTmp.find( THREEML_TOKEN_ONE_LINE_REMARKS ) != string :: npos )
    return TOKEN_TYPE_ONE_LINE_REMARKS;
    
  // Multi-line remarks (Open)
  if( strTmp.find( THREEML_TOKEN_OPEN_MULTI_LINES_REMARKS ) != string :: npos )
    return TOKEN_TYPE_OPEN_MULTI_LINES_REMARKS;

  // Multi-line remarks (Close)
  if( strTmp.find( THREEML_TOKEN_CLOSE_MULTI_LINES_REMARKS ) != string :: npos )
    return TOKEN_TYPE_CLOSE_MULTI_LINES_REMARKS;
  
  /*
   * Warning: Always the last thing  to do in this method.
   */
  trim( strTmp );

  if( strTmp.size() <= 0 )
    return TOKEN_TYPE_BLANK;

  return TOKEN_TYPE_DATA;
}

/**
  * Extract data from token based on toke type information;
  * @param type The Token type to extract data;
  * @param strToken The token containing the data;
  * @param token The reference variable to receive parsed token value;
  */
void ThreeML :: ExtractTokenValue( TokenType type, string strToken, stToken &token )  {
  
  switch( type )  {
    
    case TOKEN_TYPE_SECTION :  {
                                 size_t    nStart = strToken.find( THREEML_TOKEN_OPEN_BRACKET );
                                 size_t    nEnd   = strToken.find( THREEML_TOKEN_CLOSE_BRACKET );
                                 size_t    nLen   = ( nEnd - nStart ) - 1;
                                 
                                 /* 
                                  * Parse the section name and return
                                  * as Key/Value pair.
                                  */
                                 token.type     = DATA_TYPE_KEY_VALUE;
                                 token.strKey   = strToken.substr( nStart + 1, nLen );
                                 token.strValue = token.strKey;
                                 trim( token.strKey );
                                 trim( token.strValue );
                               }  break;
    case TOKEN_TYPE_DATA    :  {
                                 /*
                                  * Just parse the value.
                                  */
                                 if( strToken.find( THREEML_TOKEN_ASSIGN_VALUE ) == string :: npos )  {
                                   token.type     = DATA_TYPE_VALUE;
                                   token.strValue = strToken;
                                   trim( token.strValue );
                                   token.strKey.clear();
                                 }
                                 else  {
                                   size_t    nPos = strToken.find( THREEML_TOKEN_ASSIGN_VALUE );
                                   
                                   /*
                                    * Parse the key and value pair.
                                    */
                                   token.type     = DATA_TYPE_KEY_VALUE;
                                   token.strKey   = strToken.substr( 0, nPos );
                                   token.strValue = strToken.substr( nPos + 1 );
                                   trim( token.strKey );
                                   trim( token.strValue );
                                 }
                               }  break;
    case TOKEN_TYPE_OPEN_MULTI_LINES_REMARKS  :
    case TOKEN_TYPE_CLOSE_MULTI_LINES_REMARKS : { // Multi-column remarks processing
                                                  size_t    nOpenRemPos  = strToken.find( THREEML_TOKEN_OPEN_MULTI_LINES_REMARKS );
                                                  size_t    nCloseRemPos = strToken.find( THREEML_TOKEN_CLOSE_MULTI_LINES_REMARKS );

                                                  /*
                                                   * Parse the remark, retrieving the value out of the
                                                   * open and close remarks token.
                                                   */
                                                  token.type      = DATA_TYPE_VALUE;
                                                  token.strValue  = strToken.substr( 0, nOpenRemPos );
                                                  token.strValue += strToken.substr( nCloseRemPos + string( THREEML_TOKEN_CLOSE_MULTI_LINES_REMARKS ).size() );
                                                  trim( token.strValue );
                                                  token.strKey.clear();
                                                }
                                                break;
  }
}

/**
  * Filter the specified value passed by parameter.
  * Actually the filter search by following patterns:
  * Character & after numbers are removed;
  * Character . after L followed by a number;
  * @param strToken The value to be filtered;
  */
void ThreeML :: FilterValue( string &strToken )  {

  ::remove( strToken, MSD_TOKEN_SPACE );
  AmpersandFilter( strToken );
  CombineQuarterNotesFilter( strToken );
  DotFilter( strToken );
  OctaveLowerFilter( strToken );
  TildeFilter( strToken );
  InvalidVolumeFilter( strToken );

  /*
   * Apply remove instruments filter ??
   */
  if( GetRemoveInstruments() )
    InstrumentsFilter( strToken );

  ::remove( strToken, MSD_TOKEN_SPACE );
}

/**
  * Add a token to compiled MML map.
  * @param pTokenSection The token section owner of vector information;
  * @param pTokenVector The data vector of section;
  */
void ThreeML :: AddToken( stToken *pTokenSection, TokenVector *pTokenVector )  {

  /*
   * Add the last token vector to compiled MML structure.
   */
  if( pTokenVector && pTokenSection )  {
    if( ( pTokenSection -> strKey.compare( THREEML_TOKEN_SETTINGS_SECTION_KEY ) != 0 ) &&
        ( pTokenSection -> strKey.compare( THREEML_TOKEN_3MLE_EXTENSION_KEY ) != 0 ) )  {
      stringstream   strNewKey;

      /*
       * Change all section name to internal name.
       */
      strNewKey << __MSD_CHANNEL_MML_VARIABLE << ( nChannelId + 1 );
      pTokenSection -> strKey = strNewKey.str();
    }

    compiledMML.insert( make_pair( pTokenSection, pTokenVector ) );
    /*
     * Must ignore the settings key to add channels.
     */
    if( ( pTokenSection -> strKey.compare( THREEML_TOKEN_SETTINGS_SECTION_KEY ) != 0 ) &&
        ( pTokenSection -> strKey.compare( THREEML_TOKEN_3MLE_EXTENSION_KEY ) != 0 ) )
      channels.insert( make_pair( nChannelId++, pTokenSection -> strKey ) );
  }
}

/**
  * Generate output for the specified channel.
  * @param channelType The channel type to generate output;
  * @param nChannelId The channel id to generate output;
  */
void ThreeML :: MakeChannel( ChannelType channelType, int nChannelId )  {
  
  ChannelMap :: iterator itChannel;
  string                 strChannelType;
  int                    nId;
  stringstream           strCustomMMLVar;
  

  switch( channelType )  { 
    case CHANNEL_TYPE_FM  :  nId = ( nChannelId > MSD_FMR_ID ? nChannelId - 2 : nChannelId - 1 );
                             strChannelType = MSD_FM_TOKEN;  
                             break;
    case CHANNEL_TYPE_PSG :  nId = ( nChannelId + MSD_MAX_FM_CHANNELS - 2 );
                             strChannelType = MSD_PSG_TOKEN;  
                             break;
    case CHANNEL_TYPE_SCC :  nId = ( nChannelId + MSD_MAX_FM_CHANNELS + MSD_MAX_PSG_CHANNELS - 2 );
                             strChannelType = MSD_SCC_TOKEN;
                             break;
  }

  strCustomMMLVar.clear();

  /*
   * Build the custom MML variable.
   */
  for( CustomMMLMap :: iterator itItems = customMML.begin(); itItems != customMML.end(); itItems++ )  {
    ChannelSet   *pChannelSet = itItems -> second;

    if( pChannelSet -> find( nId + 1 ) != pChannelSet -> end() )  {
      stToken      *pToken = itItems -> first;

      strCustomMMLVar << pToken -> strKey
                      << MSD_TOKEN_COMMA;                      
    }
  }

  out << strChannelType;

  if( nChannelId != MSD_FMR_ID )  {
    itChannel = channels.find( nId );
    
    out << ( nChannelId > MSD_FMR_ID ? ( nChannelId - 1 ) : nChannelId )
        << MSD_TOKEN_ASSIGN_VALUE
        << strCustomMMLVar.str()
        << ( itChannel != channels.end() ? itChannel -> second : "" ) 
        << MSD_TOKEN_EOL;
  }
  else
    out << MSD_R_TOKEN
        << MSD_TOKEN_ASSIGN_VALUE
        << MSD_TOKEN_EOL;
}

/**
  * Compile a 3ML MML file to a internal MML map.
  */
bool ThreeML :: Compile( void )  {

  string          strToken;
  TokenType       tokenType;
  stToken         *pTokenSection = NULL;
  TokenVector     *pTokenVector  = NULL;
  bool            bSyntaxError   = false;


  nChannelId = 0;

  /*
   * Start parsing 3ML MML file.
   */
  while( in.good() )  {
    getline( in, strToken );

    ::charcase( strToken, UPPERCASE );

    if( GetVerbose() )  
      cout << strToken << "\n";

    /*
     * Start processing input lines checking all possible data types.
     */
    tokenType = CheckToken( strToken );

    switch( tokenType )  {
      case TOKEN_TYPE_SECTION : AddToken( pTokenSection, pTokenVector );                                 
                                pTokenVector  = new TokenVector;
                                pTokenSection = new stToken;
                                ExtractTokenValue( tokenType, strToken, *pTokenSection );
                                break;
      case TOKEN_TYPE_OPEN_MULTI_LINES_REMARKS :  // Multi-column remarks processing
      case TOKEN_TYPE_CLOSE_MULTI_LINES_REMARKS:
      case TOKEN_TYPE_DATA    : if( !pTokenVector )  {
                                  bSyntaxError = true;
                                  cout << "Syntax error. Undefined section\n";
                                }
                                else  {
                                  stToken    *pToken = new stToken;
                                  
                                  ExtractTokenValue( tokenType, strToken, *pToken );
                                  pTokenVector -> push_back( pToken );
                                }
                                break;
      case TOKEN_TYPE_UNKNOWN : cout << "Syntax error\n";  
                                break;
      case TOKEN_TYPE_ONE_LINE_REMARKS : // Blank statements
      case TOKEN_TYPE_BLANK            : break;
    }

    /*
     * Has syntax error in MML ?
     */
    if( bSyntaxError )  {
      cout << "The input MML file has errors. Unable to generate output MML file\n";
      break;
    }
  }
  
  /*
   * Add the last token vector to compiled MML structure.
   */
  if( !bSyntaxError )
    AddToken( pTokenSection, pTokenVector );

  if( bSyntaxError )
    Clear();

  in.close();

  return true;
}

/**
  * Link the compiled MML to a MSD MML file.
  */
bool ThreeML :: Link( void )  {

  /*
   * Apply pre filters processing.
   */
  if( GetFindTempo() )
    PreFilters();

  /*
   * MSD generated header file.
   */
  if( GetRemarks() )  {
    out << MSD_TOKEN_GENERATED_FILE << MSD_TOKEN_EOL;
    out << MSD_TOKEN_COPYLEFT << MSD_TOKEN_EOL;
    out << MSD_TOKEN_REMARK << MSD_TOKEN_EOL << MSD_TOKEN_EOL;

    out << MSD_TOKEN_REMARK << MSD_TOKEN_EOL;
    out << MSD_TOKEN_CHANNELS_DEFINITIONS << MSD_TOKEN_EOL;
    out << MSD_TOKEN_REMARK << MSD_TOKEN_EOL << MSD_TOKEN_EOL;
  }

  /*
   * Dump channel definitions to output file.
   */
  MakeChannel( CHANNEL_TYPE_FM, 1 );
  MakeChannel( CHANNEL_TYPE_FM, 2 );
  MakeChannel( CHANNEL_TYPE_FM, 3 );
  MakeChannel( CHANNEL_TYPE_FM, 4 );
  MakeChannel( CHANNEL_TYPE_FM, 5 );
  MakeChannel( CHANNEL_TYPE_FM, 6 );
  MakeChannel( CHANNEL_TYPE_FM, 7 );  // FMR channel
  MakeChannel( CHANNEL_TYPE_FM, 8 );
  MakeChannel( CHANNEL_TYPE_FM, 9 );
  MakeChannel( CHANNEL_TYPE_FM, 10 );
  
  MakeChannel( CHANNEL_TYPE_PSG, 1 );
  MakeChannel( CHANNEL_TYPE_PSG, 2 );
  MakeChannel( CHANNEL_TYPE_PSG, 3 );
  
  MakeChannel( CHANNEL_TYPE_SCC, 1 );
  MakeChannel( CHANNEL_TYPE_SCC, 2 );
  MakeChannel( CHANNEL_TYPE_SCC, 3 );
  MakeChannel( CHANNEL_TYPE_SCC, 4 );
  MakeChannel( CHANNEL_TYPE_SCC, 5 );

  if( GetRemarks() && ( customMML.size() > 0 ) )  {
    out << MSD_TOKEN_EOL << MSD_TOKEN_REMARK << MSD_TOKEN_EOL;
    out << MSD_TOKEN_CUSTOM_MML_DEFINITIONS << MSD_TOKEN_EOL;
    out << MSD_TOKEN_REMARK << MSD_TOKEN_EOL << MSD_TOKEN_EOL;
  }

  /*
   * Dump the custom MML to output file.
   */
  for( CustomMMLMap :: iterator itCustomMMLItems = customMML.begin(); itCustomMMLItems != customMML.end(); itCustomMMLItems++ )  {
    stringstream     strCustomMMLVar;
    stToken          *pToken = itCustomMMLItems -> first;
    
    strCustomMMLVar << pToken -> strKey
                    << MSD_TOKEN_ASSIGN_VALUE 
                    << pToken -> strValue
                    << MSD_TOKEN_EOL;
    out << strCustomMMLVar.str();
  }
  
  if( GetRemarks() )  {
    out << MSD_TOKEN_EOL << MSD_TOKEN_REMARK << MSD_TOKEN_EOL;
    out << MSD_TOKEN_MML_CONTENT << MSD_TOKEN_EOL;
    out << MSD_TOKEN_REMARK << MSD_TOKEN_EOL << MSD_TOKEN_EOL;
  }

  /*
   * Dump all channels content to output file.
   */
  for( TokenVectorMap :: iterator itItems = compiledMML.begin(); itItems != compiledMML.end(); itItems++ )  {
    TokenVector    *pTokenVector = itItems -> second;
    
    if( pTokenVector )  {

      string      strValue( "" );

      /*
       * Must ignore the settings and 3ML extension keys at link process.
       */
      if( ( itItems -> first -> strKey.compare( THREEML_TOKEN_SETTINGS_SECTION_KEY ) == 0 ) ||
          ( itItems -> first -> strKey.compare( THREEML_TOKEN_3MLE_EXTENSION_KEY ) == 0 ) )
        continue;
 
      out << itItems -> first -> strKey << "=";

      for( TokenVector :: iterator itTokenItems = pTokenVector -> begin(); itTokenItems != pTokenVector -> end(); itTokenItems++ )  {
        struct   stToken    *pToken = *itTokenItems;

        strValue += pToken -> strValue;
      }

      /*
       * Apply filter to data removing incompatible MML
       * sequences.
       */
      FilterValue( strValue );

      out << ( strValue.empty() ? __MSD_DEFAULT_TAG_FOR_EMPTY_CHANNEL : strValue );
      out << MSD_TOKEN_EOL;
    }
  }
  
  out << MSD_TOKEN_EOL;

  /*
   * MSD file terminator.
   */
  out.put( MSD_TOKEN_EOF );

  Clear();
  
  return true;
}

/**
  * Clear all class data.
  */
void ThreeML :: Clear( void )  {
  
  for( TokenVectorMap :: iterator itItems = compiledMML.begin(); itItems != compiledMML.end(); itItems++ )  {
    stToken        *pTokenSection = itItems -> first;
    TokenVector    *pTokenVector  = itItems -> second;
    
    if( pTokenVector )  {
      for( TokenVector :: iterator itTokenItems = pTokenVector -> begin(); itTokenItems != pTokenVector -> end(); itTokenItems++ )
        delete *itTokenItems;

      pTokenVector -> clear();
      delete pTokenVector;
    }

    if( pTokenSection )
      delete pTokenSection;
  }
  
  RemovelAllCustomMML();

  compiledMML.clear();
  channels.clear();
  customMML.clear();
  bTempoFound = false;
  nChannelId  = 0;
}

// Public declarations
/**
  * Constructor. Initialize all class data.
  * @param strInputFileName The input MML file to be converted;
  * @param strOutputFileName The output MSD file name;
  */
ThreeML :: ThreeML( const string strInputFileName, const string strOutputFileName )  {

  ThreeML :: strInputFileName  = strInputFileName;
  ThreeML :: strOutputFileName = strOutputFileName;
  nVarCount   = 1;
  bTempoFound = false;
  SetVerbose( true );
  SetRemarks( true );
  SetFindTempo( false );
}

/**
  * Destructor. Clear all resources allocated by MML object.
  */
ThreeML :: ~ThreeML( void )  {

  Close();
  Clear();
}

/**
  * Open the input MML and the output MSD files.
  */
bool ThreeML :: Open( void )  {
  
  // Open input MML file
  in.open( strInputFileName.c_str(), ios_base :: in );

  if( !in.is_open() )  {
    cout << "MML.Open() - Error to open MML input file [" << strInputFileName.c_str() << "]";
    return false;
  }

  out.open( strOutputFileName.c_str(), ios_base :: out );

  // Open output MSD file
  if( !out.is_open() )  {
    in.close();
    cout << "MML.Open() - Error to open MSD output file [" << strOutputFileName.c_str() << "]";
    return false;
  }

  return true;  
}

/**
  * Close the input MML and the output MSD files.
  */
bool ThreeML :: Close( void )  {

  if( in.is_open() )
    in.close();

  if( out.is_open() )
    out.close();

  return true;
}

/**
  * Add a custom MML command to a specified channel.
  * @param strCustomMML The custom MML to add to specified channel;
  * @param nChannel The Channel to add the custom MML;
  */
void ThreeML :: AddCustomMML( string strCustomMML, int nChannel )  {

  bool                      bInsertInMap = false;
  ChannelSet                *pChannelSet = NULL;
  stToken                   *pToken;

  /*
   * Try to find a ChannelSet allocated for the custom MML
   * received by function.
   */
  for( CustomMMLMap :: iterator itItems = customMML.begin(); itItems != customMML.end(); itItems++ )  {
    pToken = itItems -> first;
    
    if( pToken -> strValue.compare( strCustomMML ) == 0 )  {      
      pChannelSet = itItems -> second;
      break;
    }
  }
  
  /*
   * MML not found in list.
   */
  if( !pChannelSet )  {
    stringstream     strCustomMMLVar;
    
    strCustomMMLVar << __MSD_CUSTOM_MML_VARIABLE 
                    << nVarCount++;

    pChannelSet = new ChannelSet;
    pToken      = new stToken;

    pToken -> type     = DATA_TYPE_KEY_VALUE;
    pToken -> strKey   = strCustomMMLVar.str();
    pToken -> strValue = strCustomMML;
    bInsertInMap = true;
  }

  if( nChannel == ALL_CHANNELS )  {
    for( int nCount = 1; nCount <= MSD_MAX_CHANNELS; nCount++ )
      pChannelSet -> insert( nCount );
  }
  else
    pChannelSet -> insert( nChannel );
  
  if( bInsertInMap )
    customMML.insert( make_pair( pToken, pChannelSet ) );
}

/**
  * Remove all custom MML commands added by
  * @see AddCustomMML().
  */
void ThreeML :: RemovelAllCustomMML( void )  {

  /*
   * Clear the custom MML list.
   */
  for( CustomMMLMap :: iterator itCustomItems = customMML.begin(); itCustomItems != customMML.end(); itCustomItems++ )  {
    stToken      *pToken      = itCustomItems -> first;
    ChannelSet   *pChannelSet = itCustomItems -> second;
    
    if( pToken )
      delete pToken;

    if( pChannelSet )  {
      pChannelSet -> clear();
      delete pChannelSet;
    }
  }
  
  nVarCount = 1;
}
